@@ -14,21 +14,13 @@ from django_upgrade.tokens import update_import_modules
 
 fixer = Fixer(
     __name__,
-    min_version=(1, 11),
+    min_version=(1, 9),
 )
 
 REWRITES = {
-    "django.db.models.fields": {
-        "FieldDoesNotExist": "django.core.exceptions",
-    },
-    "django.db.models.query": {
-        "EmptyResultSet": "django.core.exceptions",
-    },
-    "django.db.models.sql": {
-        "EmptyResultSet": "django.core.exceptions",
-    },
-    "django.db.models.sql.datastructures": {
-        "EmptyResultSet": "django.core.exceptions",
+    "django.forms.forms": {
+        "pretty_name": "django.forms.utils",
+        "BoundField": "django.forms.boundfield",
     },
 }
 
