@@ -1,23 +1,28 @@
 import ast
+from collections import defaultdict
 from typing import Dict, List, Optional, Tuple
 
 from tokenize_rt import UNIMPORTANT_WS, Token, tokens_to_src
 
 # Token name aliases
-INDENT = "INDENT"
+CODE = "CODE"  # Token name meaning 'replaced by us'
+COMMENT = "COMMENT"
 DEDENT = "DEDENT"
+INDENT = "INDENT"
+LOGICAL_NEWLINE = "NEWLINE"
 NAME = "NAME"
 OP = "OP"
-LOGICAL_NEWLINE = "NEWLINE"
 PHYSICAL_NEWLINE = "NL"
-COMMENT = "COMMENT"
-CODE = "CODE"  # Token name meaning 'replaced by us'
+STRING = "STRING"
 
 
-BRACES = {"(": ")", "[": "]", "{": "}"}
+# Basic functions
 
 
 def find(tokens: List[Token], i: int, *, name: str, src: Optional[str] = None) -> int:
+    """
+    Find the next token matching name and src.
+    """
     while tokens[i].name != name or (src is not None and tokens[i].src != src):
         i += 1
     return i
@@ -26,17 +31,40 @@ def find(tokens: List[Token], i: int, *, name: str, src: Optional[str] = None) -
 def reverse_find(
     tokens: List[Token], i: int, *, name: str, src: Optional[str] = None
 ) -> int:
+    """
+    Find the previous token matching name and src.
+    """
     while tokens[i].name != name or (src is not None and tokens[i].src != src):
         i -= 1
     return i
 
 
-def find_and_replace_name(tokens: List[Token], i: int, *, name: str, new: str) -> None:
-    j = find(tokens, i, name=NAME, src=name)
-    tokens[j] = tokens[j]._replace(name="CODE", src=new)
+def consume(
+    tokens: List[Token], i: int, *, name: str, src: Optional[str] = None
+) -> int:
+    """
+    Move past any tokens matching name and src.
+    """
+    while tokens[i + 1].name == name and (src is None or tokens[i + 1].src == src):
+        i += 1
+    return i
+
+
+def reverse_consume(
+    tokens: List[Token], i: int, *, name: str, src: Optional[str] = None
+) -> int:
+    """
+    Rewind past any tokens matching name and src.
+    """
+    while tokens[i - 1].name == name and (src is None or tokens[i - 1].src == src):
+        i -= 1
+    return i
 
 
 def find_final_token(tokens: List[Token], i: int, *, node: ast.AST) -> int:
+    """
+    Find the last token corresponding to the given ast node.
+    """
     j = i
     while tokens[j].line is None or tokens[j].line < node.end_lineno:
         j += 1
@@ -48,32 +76,24 @@ def find_final_token(tokens: List[Token], i: int, *, node: ast.AST) -> int:
     return j
 
 
-def erase_node(tokens: List[Token], i: int, *, node: ast.AST) -> None:
-    j = find_final_token(tokens, i, node=node)
-    if tokens[j].name == LOGICAL_NEWLINE:  # pragma: no branch
-        j += 1
-    # if i > 0 and tokens[i - 1].name == INDENT:
-    #     i -= 1
-    del tokens[i:j]
-
-
-def consume(
-    tokens: List[Token], i: int, *, name: str, src: Optional[str] = None
-) -> int:
-    while tokens[i + 1].name == name and (src is None or tokens[i + 1].src == src):
-        i += 1
-    return i
-
-
-def reverse_consume(
-    tokens: List[Token], i: int, *, name: str, src: Optional[str] = None
-) -> int:
-    while tokens[i - 1].name == name and (src is None or tokens[i - 1].src == src):
-        i -= 1
-    return i
+def extract_indent(tokens: List[Token], i: int) -> Tuple[int, str]:
+    """
+    If the previous token is and indent, return its position and the
+    indentation string. Otherwise return the current position and "".
+    """
+    j = i
+    if j > 0 and tokens[j - 1].name == INDENT:
+        j -= 1
+        indent = tokens[j].src
+    else:
+        indent = ""
+    return (j, indent)
 
 
 def alone_on_line(tokens: List[Token], start_idx: int, end_idx: int) -> bool:
+    """
+    Return if the given set of tokens is on its own physical line.
+    """
     return (
         tokens[start_idx - 2].name == PHYSICAL_NEWLINE
         and tokens[start_idx - 1].name == UNIMPORTANT_WS
@@ -81,25 +101,21 @@ def alone_on_line(tokens: List[Token], start_idx: int, end_idx: int) -> bool:
     )
 
 
-def insert(tokens: List[Token], i: int, *, new_src: str) -> None:
-    tokens.insert(i, Token(CODE, new_src))
+# More complex mini-parser functions
 
 
-def insert_after(
-    tokens: List[Token], i: int, *, name: str, src: Optional[str] = None, new_src: str
-) -> None:
-    j = find(tokens, i, name=name, src=src)
-    tokens.insert(j + 1, Token(CODE, new_src))
-
-
-def replace(tokens: List[Token], i: int, *, src: str) -> None:
-    tokens[i] = tokens[i]._replace(name=CODE, src=src)
+BRACES = {"(": ")", "[": "]", "{": "}"}
 
 
 def parse_call_args(
     tokens: List[Token],
     i: int,
 ) -> Tuple[List[Tuple[int, int]], int]:
+    """
+    Given the index of the opening bracket of a function call, step through
+    and parse its arguments into a list of tuples of start, end indices.
+    Return this list plus the position of the token after.
+    """
     args = []
     stack = [i]
     i += 1
@@ -124,13 +140,51 @@ def parse_call_args(
     return args, i
 
 
-def replace_arguments(
+# Rewriting functions
+
+
+def insert(tokens: List[Token], i: int, *, new_src: str) -> None:
+    """
+    Insert a generated token with the given new source.
+    """
+    tokens.insert(i, Token(CODE, new_src))
+
+
+def replace(tokens: List[Token], i: int, *, src: str) -> None:
+    """
+    Replace the token at position i with a generated token with the given new
+    source.
+    """
+    tokens[i] = tokens[i]._replace(name=CODE, src=src)
+
+
+def erase_node(tokens: List[Token], i: int, *, node: ast.AST) -> None:
+    """
+    Erase all tokens corresponding to the given ast node.
+    """
+    j = find_final_token(tokens, i, node=node)
+    if tokens[j].name == LOGICAL_NEWLINE:  # pragma: no branch
+        j += 1
+    i, _ = extract_indent(tokens, i)
+    del tokens[i:j]
+
+
+def find_and_replace_name(tokens: List[Token], i: int, *, name: str, new: str) -> None:
+    j = find(tokens, i, name=NAME, src=name)
+    tokens[j] = tokens[j]._replace(name="CODE", src=new)
+
+
+def replace_argument_names(
     tokens: List[Token],
     i: int,
     *,
     node: ast.Call,
     arg_map: Dict[str, str],
 ) -> None:
+    """
+    Update an ast.Call node’s keyword argument names, where arg_map maps old to
+    new names.
+    """
     j = find(tokens, i, name=OP, src="(")
     func_args, _ = parse_call_args(tokens, j)
     kwarg_func_args = func_args[len(node.args) :]
@@ -146,13 +200,17 @@ def replace_arguments(
                 raise AssertionError(f"{keyword.arg} argument not found")
 
 
-def update_imports(
+def update_import_names(
     tokens: List[Token],
     i: int,
     *,
     node: ast.ImportFrom,
     name_map: Dict[str, str],
 ) -> None:
+    """
+    Replace an ast.ImportFrom node’s imported names, where name_map maps old to
+    new names. If a new name entry is the empty string, remove the import.
+    """
     j = find(tokens, i, name=NAME, src="from")
     j = find(tokens, j, name=NAME, src="import")
 
@@ -215,3 +273,32 @@ def update_imports(
     else:
         for start_idx, end_idx, replacement in reversed(replacements):
             tokens[start_idx : end_idx + 1] = replacement
+
+
+def update_import_modules(
+    tokens: List[Token],
+    i: int,
+    *,
+    node: ast.ImportFrom,
+    module_rewrites: Dict[str, str],
+) -> None:
+    """
+    Replace import names from an ast.ImportFrom with new import statements from
+    elsewhere. rewrites should map import names to the new modules they should
+    be imported from.
+    """
+    assert node.module is not None
+    imports_to_add = defaultdict(list)
+    name_map = {}
+    for alias in node.names:
+        name = alias.name
+        if name in module_rewrites:
+            name_map[name] = ""
+            new_name = f"{name} as {alias.asname}" if alias.asname else name
+            imports_to_add[module_rewrites[name]].append(new_name)
+
+    j, indent = extract_indent(tokens, i)
+    update_import_names(tokens, i, node=node, name_map=name_map)
+    for module, names in reversed(imports_to_add.items()):
+        joined_names = ", ".join(sorted(names))
+        insert(tokens, j, new_src=f"{indent}from {module} import {joined_names}\n")
